!!$=head1 NAME
!!$
!!$Born modeling Adjoint - Acoustic RTM operator for non-constant density.
!!$
!!$=head1 SYNOPSIS
!!$
!!$Bornfast_acoustic_adj.x <bulk.H rho=rho.H deltadata=deltadata.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefields=wavefields.H wavefieldr=wavefieldr.H deltarho=deltarho.H >deltak.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation Adjoint Born modeling for full acoustic approximation with absorbing boundary (i.e., RTM). It takes the source and receiver positions from the parameters file created by Velpar.x and propagates the wavelet into
!!$the velocity model, the perturbation in the data and return the perturbed model. !NOTE= it assumes that we can keep in memory the source wavefield at source-wavelet sampling!
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<bulk.H> - Bulk modulus model in SEP3D format (theoretically [GPa]).
!!$
!!$=item from par
!!$
!!$B<filename> I<rho> - density model in SEP3D format (theoretically [Kg/m^3]).
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<deltadata> - data perturbation to be back-propagated.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefields> - [wavefields.H] File where to save source wavefield if wantwavefield=1
!!$
!!$B<filename> I<wavefieldr> - [wavefieldr.H] File where to save receiver wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional)
!!$
!!$B<integer> I<threads> - [8] Number of threads to use during propagation.
!!$
!!$B<filename> I<deltak> - Bulk modulus perturbation in SEP format.
!!$
!!$B<filename> I<deltarho> - Density perturbation in SEP format.
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Bornfast_acoustic_fw
  !Used modules 
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use Add_source2D
  use Add_source3D
  use resample 
  use fourier
  use Sou_scale_acoustic
  use Prop_acoustic_step
  use weight_sec
  use S_mod
  use sec_deriv_field
  use D8pt_mod


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: bulk, rho  !bulk modulus and density models 
  real, dimension(:), allocatable						:: wavelet, scaled_wavelet !Source wavelet 
  integer												:: wantwavefield, num_threads
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold
  real, dimension(:,:,:), pointer						:: bulk_pad, rho_pad !Padded models
  real													:: dt !Temporal sampling of the propagator
  real													:: maxvel, minvel !max and min velocity
  real, dimension(:), allocatable						:: wavelet_fine !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the models 
  integer												:: is, ig, it !Cycles variables
  integer												:: wc, srite, rc, sreed, sk, sseek
  integer												:: desample
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr
  integer												:: count1, count2,ierr
  !Born modeling variables
  integer												:: it2, itw, icount
  real, dimension(:,:,:), allocatable					:: pertk, pertrho   !K and rho perturbation
  real, dimension(:,:,:), pointer						:: pertk_pad, pertrho_pad   !K and rho perturbation padded
  real, dimension(:,:,:), pointer						:: tmp_vel	!necessary for weight_sec
  real, dimension(:,:,:,:),pointer						:: source_wavefield !Source wavefield at coarse scale
  real*8, dimension(:,:), allocatable					:: interp_filter
  real, dimension(:,:,:), allocatable, target			:: sec_source_left, sec_source_right
  real, dimension(:,:,:), pointer						:: psec_source_left, psec_source_right, ptemp
  real, dimension(:,:,:), allocatable					:: sec_scale
  real, dimension(:,:,:), allocatable					:: pertrhoz, pertrhox, pertrhoy !Staggered density perturbation
  real, dimension(:,:,:), pointer						:: pnewpert, pcurpert, poldpert
  real, dimension(:,:,:), pointer						:: tmp_scat, scat_sou

  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
  
  write(0,*)""
  write(0,*)""
  write(0,*)"##############################################################################"
  write(0,*)"#################### Full Acoustic 2D/3D wave propagator #####################"
  write(0,*)"####################      Adjoint Born modeling (RTM)    #####################" 
  write(0,*)"##############################################################################"
  write(0,*)""
  write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("dt",dt,0.)
  !Setting number of threads
  call from_par("threads",num_threads,8)
  call omp_set_num_threads(num_threads)
  
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy);
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the bulk modulus
  allocate(bulk(nz,nx,ny))
  bulk(:,:,:) = 0.
  rc=sreed("in" , bulk , nz*nx*ny*4)
  bulk(:,:,:) = bulk(:,:,:)*1e9 ! Conversion from GPa to Pa
  !Reading the density model
  allocate(rho(nz,nx,ny))
  rho(:,:,:) = 0.
  rc=sreed("rho" , rho , nz*nx*ny*4)
  !Computing max and min velocities
  minvel = minval(sqrt(bulk/rho))
  maxvel = maxval(sqrt(bulk/rho))
  
  !Adding absorbing boundaries to the velocity model and perturbation
  	!Adding absorbing boundaries to the bulk model
  	call Pad_velocity(bulk,bulk_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	!Adding absorbing boundaries to the density model and perturbations
  	call Pad_perturbation(rho,rho_pad,nv_pad)
  	allocate(pertk_pad(nz,nx,ny));pertk_pad(:,:,:) = 0.
  	allocate(pertrho_pad(nz,nx,ny));pertrho_pad(:,:,:) = 0.
  
  write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  write(0,*) "	Number of samples for 50 Hz wave = ", (minvel/(50.*max(dx,dy,dz)))
  write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  write(0,*)"	Min vel [m/s]:",minvel
  write(0,*)"	Max vel [m/s]:",maxvel
  write(0,*) ""
  
  

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  	 desample = 1
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	call resample_init(nw, nt)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)
  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_acoustic_step_init(nz, nx, ny, nt, nv_pad, bulk_pad, rho_pad, dt, dz, dx, dy, twoD)
  allocate(tmp_vel(nz,nx,ny));tmp_vel = 0.
  tmp_vel = sqrt(bulk_pad/rho_pad)
  call weight_sec_init(nz,nx,ny,dt,dz,dx,dy,nv_pad,tmp_vel,twoD)
  deallocate(tmp_vel) 
  call S_init(nz,nx,ny,twoD)
  call sec_deriv_field_init(nw, dw)
  call D8pt_init(nz,nx,ny,dz,dx,dy,twoD)
  
  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  old = 0.; cur = 0.
  pold => old
  pcur => cur
  pnew => old
  allocate(sec_scale(nz,nx,ny))
  sec_scale(:,:,:) = dt*dt*bulk_pad !Check this when comparing trueDdata with Born modeled data
  allocate(sec_source_left(nz,nx,ny),sec_source_right(nz,nx,ny))
  psec_source_left=>sec_source_left
  psec_source_right=>sec_source_right
  !allocating memory for source wavefield 
  allocate(source_wavefield(nz,nx,ny,nw))
  source_wavefield = 0.
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))
  !Staggered density perturbation
  allocate(pertrhoz(nz,nx,ny), pertrhox(nz,nx,ny))
  if(.not. twoD) then
	allocate(pertrhoy(nz,nx,ny))
  endif
  allocate(scat_sou(nz,nx,ny),tmp_scat(nz,nx,ny))
  
  write(0,*) "Dimensions:"
  write(0,'(A7,I6)') " nz = ",nz
  write(0,'(A7,I6)') " nx = ",nx
  write(0,'(A7,I6)') " ny = ",ny
  write(0,'(A7,I6)') " nt = ",nt
  write(0,'(A7,I6)') " ns = ",ns
  write(0,'(A7,I6)') " ng = ",ng
  write(0,*) ""
  
  if (wantwavefield) then
	write(0,*)"#Program will be writing wavefield as requested#"
	do is = 1,ns
		do it = 1, nw
			wc=srite("wavefieldr",pcur,nz*nx*ny*4)
		enddo
	enddo
  endif
  
  
  do is = 1,ns
  	!moving the file writing position with sseek
  	if (wantwavefield) sk = sseek("wavefieldr",4*nz*nx*ny*nw*(is),0)
  	old = 0.; cur = 0.
  	sec_source_left = 0.; sec_source_right = 0.
  	scat_sou = 0.; tmp_scat = 0.
  	source_wavefield = 0.
	write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	!Scaling the source
	scaled_wavelet = wavelet_fine
	call Sou_scale_acoustic_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_acoustic_op(scaled_wavelet,bulk_pad/rho_pad) !Adjoint = Forward (input-output same variable)
	call system_clock(count_rate=cr)
	call system_clock(count1)
	write(0,*)"Propagating source wavefield..."
	if (twoD) then
		call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
	else
		call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
	endif
	do itw = 1, nw-1 !Forward
		it = (itw-1)*desample
		!Propagate source field
		do it2 = 1, desample
			call Prop_acoustic_step_op(.false., pold, pcur, pnew)
			!Adding source (perturbed and source wavefield)
			if (twoD) then
				call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			else
				call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			endif	
			!write(0,*)"source it",it+it2
			!Apply adjoint(fine to coarse) of linear interpolation
			source_wavefield(:,:,:,itw) = source_wavefield(:,:,:,itw) + pcur*interp_filter(it2,1)
			source_wavefield(:,:,:,itw+1) = source_wavefield(:,:,:,itw+1) + pcur*interp_filter(it2,2)	
			!Swapping pointers
			pold => pcur 
			pcur => pnew
			pnew => pold
		enddo
		!saving source wavefield
		if (wantwavefield) then
			wc=srite("wavefields",source_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		
		if (mod(itw,(nw/10))==0) then
        	write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	
	enddo
		!Last step of the source wavefield just saving source wavefield
		source_wavefield(:,:,:,nw) = source_wavefield(:,:,:,nw) + pcur*interp_filter(desample+1,2)
		if (wantwavefield) then
			wc=srite("wavefields",source_wavefield(:,:,:,nw),nz*nx*ny*4)
		endif
		write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	write(0,'(a)',advance="yes")""
	write(0,*)"End Propagation source wavefield"

!	write(0,*)"Reading source wavefield"
!	rc=sreed("source_wavefield",source_wavefield,nz*nx*ny*nw*4)
!	wc=srite("wavefields",source_wavefield,nz*nx*ny*nw*4)
	
	!Receiver wavefield and imaging condition
	write(0,*)"Propagating receiver wavefield..."
	icount = 0
	old = 0.; cur = 0.
	poldpert => old
  	pcurpert => cur
  	pnewpert => old
	!Creating "adjoint" source at the receivers
	if (dw .le. dt) then
		do ig = 1,ng
			rc=sreed("deltadata",recfield,nw*4)
			recfield_fine(:,ig) = recfield 
		enddo
	else
		do ig = 1,ng
			rc=sreed("deltadata",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_fine(:,ig))
		enddo
	endif
	!zero-out the model perturbations
	pertk_pad(:,:,:) = 0.
	pertrhoz(:,:,:) = 0.
	pertrhox(:,:,:) = 0.
	if(.not. twoD) then
		pertrhoy(:,:,:) = 0.
	endif
	do itw = nw, 2, -1 !Adjoint 
		!Propagate scattered wavefield
		it = (itw-2)*desample
		do it2 = desample+1, 2, -1
			call Prop_acoustic_step_op(.true., poldpert, pcurpert, pnewpert)
			!Adding receiver deltadata 
			if (twoD) then
				do ig = 1, ng
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.true.,.true.,pnewpert,recfield_fine(it+it2,ig))
				enddo
			else
				do ig = 1,ng
					call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
					call Add_source3D_op(.true.,.true.,pnewpert,recfield_fine(it+it2,ig))
				enddo
			endif
			!Apply adjoint(fine to coarse) of linear interpolation
			!write(0,*)"adjoint pert it",it+it2
			psec_source_left = psec_source_left + pcurpert*interp_filter(it2,1)
			psec_source_right = psec_source_right + pcurpert*interp_filter(it2,2)
			
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		
		!Scale secondary source (i.e., back-propagated data)
		call weight_sec_op(.true., .false., scat_sou(:,:,:), psec_source_right(:,:,:)*sec_scale(:,:,:))
		!saving source wavefield
		if (wantwavefield) then
			sk = sseek("wavefieldr",-4*nz*nx*ny,1)
			wc=srite("wavefieldr",scat_sou(:,:,:),nz*nx*ny*4)
			sk = sseek("wavefieldr",-4*nz*nx*ny,1)
		endif
		!Second order derivative of the back-propagated data
  		call sec_deriv_field_op(source_wavefield, tmp_scat, itw)
  		!Apply imaging condition bulk modulus
		pertk_pad(:,:,:) =  pertk_pad(:,:,:) - tmp_scat(:,:,:)*scat_sou(:,:,:)
		!Apply imaging condition density perturbations
		ptemp => psec_source_right
  		!z-axis scattering 
  		call Dpz_op(.false.,.false.,source_wavefield(:,:,:,itw),tmp_scat(:,:,:))
  		call Dmz_op(.true.,.false.,ptemp(:,:,:),scat_sou(:,:,:))
  		pertrhoz(:,:,:) = pertrhoz(:,:,:) + ptemp(:,:,:)*tmp_scat(:,:,:)
  		!x-axis scattering 
  		call Dpx_op(.false.,.false.,source_wavefield(:,:,:,itw),tmp_scat(:,:,:))
  		call Dmx_op(.true.,.false.,ptemp(:,:,:),scat_sou(:,:,:))
  		pertrhox(:,:,:) = pertrhox(:,:,:) + ptemp(:,:,:)*tmp_scat(:,:,:)
  		!y-axis scattering (3D ONLY) 
  		if(.not. twoD) then
  			call Dpy_op(.false.,.false.,source_wavefield(:,:,:,itw),tmp_scat(:,:,:))
  			call Dmy_op(.true.,.false.,ptemp(:,:,:),scat_sou(:,:,:))
  			pertrhoy(:,:,:) = pertrhoy(:,:,:) + ptemp(:,:,:)*tmp_scat(:,:,:)
  		endif
  		
		icount = icount + 1
		if (mod(icount,(nw/10))==0) then
	        	write(0,'(I3,A4)',advance="no")nint((icount)*100./nw),"%..."
	        	close(0)
		endif	
		
		!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		psec_source_left = 0.
	enddo
	!Last step (already propagated)
	itw = 1
	psec_source_right = psec_source_right + pcurpert*interp_filter(1,2)
		!Scale secondary source (i.e., back-propagated data)
		call weight_sec_op(.true., .false., scat_sou(:,:,:), psec_source_right(:,:,:)*sec_scale(:,:,:))
		!saving source wavefield
		if (wantwavefield) then
			sk = sseek("wavefieldr",-4*nz*nx*ny,1)
			wc=srite("wavefieldr",scat_sou(:,:,:),nz*nx*ny*4)
			sk = sseek("wavefieldr",-4*nz*nx*ny,1)
		endif
		!Second order derivative of the source wavefield
  		call sec_deriv_field_op(source_wavefield, tmp_scat, itw)
  		!Apply imaging condition bulk modulus
		pertk_pad(:,:,:) =  pertk_pad(:,:,:) - tmp_scat(:,:,:)*scat_sou(:,:,:)
		!Apply imaging condition density perturbations
		ptemp => psec_source_right
  		!z-axis scattering 
  		call Dpz_op(.false.,.false.,source_wavefield(:,:,:,itw),tmp_scat(:,:,:))
  		call Dmz_op(.true.,.false.,ptemp(:,:,:),scat_sou(:,:,:))
  		pertrhoz(:,:,:) = pertrhoz(:,:,:) + ptemp(:,:,:)*tmp_scat(:,:,:)
  		!x-axis scattering 
  		call Dpx_op(.false.,.false.,source_wavefield(:,:,:,itw),tmp_scat(:,:,:))
  		call Dmx_op(.true.,.false.,ptemp(:,:,:),scat_sou(:,:,:))
  		pertrhox(:,:,:) = pertrhox(:,:,:) + ptemp(:,:,:)*tmp_scat(:,:,:)
  		!y-axis scattering (3D ONLY) 
  		if(.not. twoD) then
  			call Dpy_op(.false.,.false.,source_wavefield(:,:,:,itw),tmp_scat(:,:,:))
  			call Dmy_op(.true.,.false.,ptemp(:,:,:),scat_sou(:,:,:))
  			pertrhoy(:,:,:) = pertrhoy(:,:,:) + ptemp(:,:,:)*tmp_scat(:,:,:)
  		endif

		write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	write(0,'(a)',advance="yes")""
	call system_clock(count2)
	write(0,*) "Propagation time: ",(count2-count1)/cr
	write(0,'(A44,I5)') " Writing model for source n°: ", is
	call Sz_op(.true., .false., pertrho_pad, pertrhoz)
  	call Sx_op(.true., .true., pertrho_pad, pertrhox)
  	!Staggering on y if not twoD problem
  	if(.not. twoD) then
  		allocate(pertrhoy(nz,nx,ny))
  		call Sy_op(.true., .true., pertrho_pad, pertrhoy)
  	endif
  	!compute bulk and density perturbations from model perturbations 
  	pertk = pertk*1e9 ! Conversion from GPa to Pa
  	pertk_pad(:,:,:) = -pertk_pad(:,:,:)/(bulk_pad(:,:,:)*bulk_pad(:,:,:))
  	pertrho_pad(:,:,:) = -pertrho_pad(:,:,:)/(rho_pad(:,:,:)*rho_pad(:,:,:))
	!Write model for is source
	wc=srite("out",pertk_pad(:,:,:),nz*nx*ny*4)
	wc=srite("deltarho",pertrho_pad(:,:,:),nz*nx*ny*4)
   	write(0,*) ""
  enddo
  call Prop_acoustic_step_close()
  call weight_sec_close()
  
  !Writing information to headers bulk modulus
  call to_history("n1",nz)
  call to_history("o1",oz)
  call to_history("d1",dz)
  call to_history("label1","z [m]")
  call to_history("n2",nx)
  call to_history("o2",ox)
  call to_history("d2",dx)
  call to_history("label2","x [m]")
  call to_history("n3",ny)
  call to_history("o3",oy)
  call to_history("d3",dy)
  call to_history("label3","y [m]")
  call to_history("n4",ns)
  call to_history("o4",1.)
  call to_history("d4",1.)
  call to_history("label4","Source number")
  !Writing information to headers density
  call to_history("n1",nz,"deltarho")
  call to_history("o1",oz,"deltarho")
  call to_history("d1",dz,"deltarho")
  call to_history("label1","z [m]","deltarho")
  call to_history("n2",nx,"deltarho")
  call to_history("o2",ox,"deltarho")
  call to_history("d2",dx,"deltarho")
  call to_history("label2","x [m]","deltarho")
  call to_history("n3",ny,"deltarho")
  call to_history("o3",oy,"deltarho")
  call to_history("d3",dy,"deltarho")
  call to_history("label3","y [m]","deltarho")
  call to_history("n4",ns,"deltarho")
  call to_history("o4",1.,"deltarho")
  call to_history("d4",1.,"deltarho")
  call to_history("label4","Source number","deltarho")
  if (wantwavefield) then
  
  	if(twoD) then
		call to_history("n1",nz,"wavefields")
  		call to_history("o1",oz,"wavefields")
  		call to_history("d1",dz,"wavefields")
  		call to_history("label1","z [m]","wavefields")
		call to_history("n2",nx,"wavefields")
	  	call to_history("o2",ox,"wavefields")
  		call to_history("d2",dx,"wavefields")
  		call to_history("label2","x [m]","wavefields")
		call to_history("n3",nt/desample+1,"wavefields")
  		call to_history("o3",ow,"wavefields")
 	 	call to_history("d3",dt*desample,"wavefields")
  		call to_history("label3","Time [s]","wavefields")
  		call to_history("n4",ns,"wavefields")
  		call to_history("o4",1.,"wavefields")
  		call to_history("d4",1.,"wavefields")
  		call to_history("label4","Source number","wavefields")  
  		
  		call to_history("n1",nz,"wavefieldr")
  		call to_history("o1",oz,"wavefieldr")
  		call to_history("d1",dz,"wavefieldr")
  		call to_history("label1","z [m]","wavefieldr")
		call to_history("n2",nx,"wavefieldr")
	  	call to_history("o2",ox,"wavefieldr")
  		call to_history("d2",dx,"wavefieldr")
  		call to_history("label2","x [m]","wavefieldr")
		call to_history("n3",nt/desample+1,"wavefieldr")
  		call to_history("o3",ow,"wavefieldr")
 	 	call to_history("d3",dt*desample,"wavefieldr")
  		call to_history("label3","Time [s]","wavefieldr")
  		call to_history("n4",ns,"wavefieldr")
  		call to_history("o4",1.,"wavefieldr")
  		call to_history("d4",1.,"wavefieldr")
  		call to_history("label4","Source number","wavefieldr")  		
  	else
		call to_history("n1",nz,"wavefields")
  		call to_history("o1",oz,"wavefields")
  		call to_history("d1",dz,"wavefields")
  		call to_history("label1","z [m]","wavefields")
		call to_history("n2",nx,"wavefields")
	  	call to_history("o2",ox,"wavefields")
  		call to_history("d2",dx,"wavefields")
  		call to_history("label2","x [m]","wavefields")
		call to_history("n3",ny,"wavefields")
	  	call to_history("o3",oy,"wavefields")
	  	call to_history("d3",dy,"wavefields")
  		call to_history("label3","y [m]","wavefields")
		call to_history("n4",nt/desample+1,"wavefields")
  		call to_history("o4",ow,"wavefields")
 	 	call to_history("d4",dt*desample,"wavefields")
  		call to_history("label4","Time [s]","wavefields")
  		call to_history("n5",ns,"wavefields")
  		call to_history("o5",1.,"wavefields")
  		call to_history("d5",1.,"wavefields")
  		call to_history("label5","Source number","wavefields")
  		
  		call to_history("n1",nz,"wavefieldr")
  		call to_history("o1",oz,"wavefieldr")
  		call to_history("d1",dz,"wavefieldr")
  		call to_history("label1","z [m]","wavefieldr")
		call to_history("n2",nx,"wavefieldr")
	  	call to_history("o2",ox,"wavefieldr")
  		call to_history("d2",dx,"wavefieldr")
  		call to_history("label2","x [m]","wavefieldr")
		call to_history("n3",ny,"wavefieldr")
	  	call to_history("o3",oy,"wavefieldr")
	  	call to_history("d3",dy,"wavefieldr")
  		call to_history("label3","y [m]","wavefieldr")
		call to_history("n4",nt/desample+1,"wavefieldr")
  		call to_history("o4",ow,"wavefieldr")
 	 	call to_history("d4",dt*desample,"wavefieldr")
  		call to_history("label4","Time [s]","wavefieldr")
  		call to_history("n5",ns,"wavefieldr")
  		call to_history("o5",1.,"wavefieldr")
  		call to_history("d5",1.,"wavefieldr")
  		call to_history("label5","Source number","wavefieldr")
  	endif
  endif
  write(0,*) "End propagation"
  

  !Deallocating memory
  deallocate(tmp_scat,scat_sou)
  if(allocated(pertrhoy))deallocate(pertrhoy)
  deallocate(pertrhox,pertrhoz)
  deallocate(interp_filter,source_wavefield)
  deallocate(cur,old)
  deallocate(sec_source_right,sec_source_left)
  deallocate(sec_scale)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(pertrho_pad,pertk_pad,rho_pad,bulk_pad)
  deallocate(wavelet)
  call sep_close()
  call exit (0)

end program Bornfast_acoustic_fw