!!$=head1 NAME
!!$
!!$Born modeling Forward - Acoustic Born modeling for non-constant density.
!!$
!!$=head1 SYNOPSIS
!!$
!!$Bornfast_acoustic_fw.x <bulk.H rho=rho.H deltak=deltak.H deltarho=deltarho.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefields=wavefields.H wavefield=wavefield.H >data.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation Born modeling for full acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameters file created by Velpar.x and the perturbations in the model parameters, propagates the wavelet into
!!$the bulk and rho models, and returns the perturbed data in data.H
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<bulk.H> - Bulk modulus model in SEP format (theoretically [GPa]).
!!$
!!$=item from par
!!$
!!$B<filename> I<rho> - density model in SEP format (theoretically [Kg/m^3]).
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<deltak> - Bulk modulus perturbation in SEP format.
!!$
!!$B<filename> I<deltarho> - Density perturbation in SEP format.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save perturbed wavefield if wantwavefield=1
!!$
!!$B<filename> I<wavefields> - [wavefields.H] File where to save source wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional)
!!$
!!$B<integer> I<padded> - [0] The model perturbation was already padded (100 elements)
!!$
!!$B<integer> I<threads> - [8] Number of threads to use during propagation.
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Bornfast_acoustic_fw
  !Used modules 
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use Add_source2D
  use Add_source3D
  use resample 
  use fourier
  use Sou_scale_acoustic
  use Prop_acoustic_step
  use weight_sec
  use S_mod
  use sec_deriv_field
  use D8pt_mod


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: bulk, rho  !bulk modulus and density models 
  real, dimension(:), allocatable						:: wavelet, scaled_wavelet !Source wavelet 
  integer												:: wantwavefield, num_threads
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold
  real, dimension(:,:,:), pointer						:: bulk_pad, rho_pad !Padded models
  real													:: dt !Temporal sampling of the propagator
  real													:: maxvel, minvel !max and min velocity
  real, dimension(:), allocatable						:: wavelet_fine !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the models 
  integer												:: is, ig, it !Cycles variables
  integer												:: wc, srite, rc, sreed
  integer												:: desample
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr
  integer												:: count1, count2,ierr
  !Born modeling variables
  integer												:: padded, it2, itw
  real, dimension(:,:,:), allocatable					:: pertk, pertrho   !K and rho perturbation
  real, dimension(:,:,:), pointer						:: pertk_pad, pertrho_pad   !K and rho perturbation padded
  real, dimension(:,:,:), pointer						:: tmp_vel	!necessary for weight_sec
  real, dimension(:,:,:,:),pointer						:: source_wavefield !Source wavefield at coarse scale
  real*8, dimension(:,:), allocatable					:: interp_filter
  real, dimension(:,:,:), allocatable, target			:: sec_source_left, sec_source_right
  real, dimension(:,:,:), pointer						:: psec_source_left, psec_source_right, ptemp
  real, dimension(:,:,:), allocatable					:: sec_scale
  real, dimension(:,:,:), allocatable					:: pertrhoz, pertrhox, pertrhoy !Staggered density perturbation
  real, dimension(:,:,:), pointer						:: pnewpert, pcurpert, poldpert
  real, dimension(:,:,:), pointer						:: tmp_scat, scat_sou
  
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
  
  write(0,*)""
  write(0,*)""
  write(0,*)"##############################################################################"
  write(0,*)"#################### Full Acoustic 2D/3D wave propagator #####################"
  write(0,*)"####################         	  Born modeling           #####################" 
  write(0,*)"##############################################################################"
  write(0,*)""
  write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("dt",dt,0.)
  call from_par ("padded",padded,0)
  !Setting number of threads
  call from_par("threads",num_threads,8)
  call omp_set_num_threads(num_threads)
  
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy);
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the bulk modulus
  allocate(bulk(nz,nx,ny))
  bulk(:,:,:) = 0.
  rc=sreed("in" , bulk , nz*nx*ny*4)
  bulk(:,:,:) = bulk(:,:,:)*1e9 ! Conversion from GPa to Pa
  !Reading the density model
  allocate(rho(nz,nx,ny))
  rho(:,:,:) = 0.
  rc=sreed("rho" , rho , nz*nx*ny*4)
  !Computing max and min velocities
  minvel = minval(sqrt(bulk/rho))
  maxvel = maxval(sqrt(bulk/rho))
  
  !Adding absorbing boundaries to the velocity model and perturbation
  if (padded .eq. 0) then
  	!Not padded
  	allocate(pertk(nz,nx,ny));pertk(:,:,:) = 0.
  	allocate(pertrho(nz,nx,ny));pertrho(:,:,:) = 0.
  	rc=sreed("deltak" , pertk , nz*nx*ny*4)
  	rc=sreed("deltarho" , pertrho , nz*nx*ny*4)
  	pertk = pertk*1e9 ! Conversion from GPa to Pa
  	!Adding absorbing boundaries to the bulk model
  	call Pad_velocity(bulk,bulk_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	!Adding absorbing boundaries to the density model and perturbations
  	call Pad_perturbation(rho,rho_pad,nv_pad)
  	call Pad_perturbation(pertk,pertk_pad,nv_pad)
  	call Pad_perturbation(pertrho,pertrho_pad,nv_pad)
  	!compute model perturbations from bulk and density perturbations
  	pertk_pad(:,:,:) = -pertk_pad(:,:,:)/(bulk_pad(:,:,:)*bulk_pad(:,:,:))
  	pertrho_pad(:,:,:) = -pertrho_pad(:,:,:)/(rho_pad(:,:,:)*rho_pad(:,:,:))  	
  else
  	!Padded
  	!Adding absorbing boundaries to the bulk model
  	call Pad_velocity(bulk,bulk_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
 	!Adding absorbing boundaries to the density model
  	call Pad_perturbation(rho,rho_pad,nv_pad)
  	allocate(pertk_pad(nz,nx,ny));pertk_pad(:,:,:) = 0.
  	allocate(pertrho_pad(nz,nx,ny));pertrho_pad(:,:,:) = 0.
  	rc=sreed("deltak" , pertk_pad , nz*nx*ny*4)
  	rc=sreed("deltarho" , pertrho_pad , nz*nx*ny*4)
  	pertk = pertk*1e9 ! Conversion from GPa to Pa
  	!compute model perturbation from bulk and density perturbation
  	pertk_pad(:,:,:) = -pertk_pad(:,:,:)/(bulk_pad(:,:,:)*bulk_pad(:,:,:))
  	pertrho_pad(:,:,:) = -pertrho_pad(:,:,:)/(rho_pad(:,:,:)*rho_pad(:,:,:))
  endif
  
  write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  write(0,*) "	Number of samples for 50 Hz wave = ", (minvel/(50.*max(dx,dy,dz)))
  write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  write(0,*)"	Min vel [m/s]:",minvel
  write(0,*)"	Max vel [m/s]:",maxvel
  write(0,*) ""
  
  

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  	 desample = 1
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	call resample_init(nw, nt)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)
  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_acoustic_step_init(nz, nx, ny, nt, nv_pad, bulk_pad, rho_pad, dt, dz, dx, dy, twoD)
  allocate(tmp_vel(nz,nx,ny));tmp_vel = 0.
  tmp_vel = sqrt(bulk_pad/rho_pad)
  call weight_sec_init(nz,nx,ny,dt,dz,dx,dy,nv_pad,tmp_vel,twoD)
  deallocate(tmp_vel) 
  call S_init(nz,nx,ny,twoD)
  call sec_deriv_field_init(nw, dw)
  call D8pt_init(nz,nx,ny,dz,dx,dy,twoD)
  
  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  old = 0.; cur = 0.
  pold => old
  pcur => cur
  pnew => old
  allocate(sec_scale(nz,nx,ny))
  sec_scale(:,:,:) = dt*dt*bulk_pad !Check this when comparing trueDdata with Born modeled data
  allocate(sec_source_left(nz,nx,ny),sec_source_right(nz,nx,ny))
  psec_source_left=>sec_source_left
  psec_source_right=>sec_source_right
  !allocating memory for source wavefield 
  allocate(source_wavefield(nz,nx,ny,nw))
  source_wavefield = 0.
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))
  allocate(scat_sou(nz,nx,ny),tmp_scat(nz,nx,ny))
  
  write(0,*) "Dimensions:"
  write(0,'(A7,I6)') " nz = ",nz
  write(0,'(A7,I6)') " nx = ",nx
  write(0,'(A7,I6)') " ny = ",ny
  write(0,'(A7,I6)') " nt = ",nt
  write(0,'(A7,I6)') " ns = ",ns
  write(0,'(A7,I6)') " ng = ",ng
  write(0,*) ""
  
  !Staggering density perturbation
  allocate(pertrhoz(nz,nx,ny), pertrhox(nz,nx,ny))
  call Sz_op(.false., .false., pertrho_pad, pertrhoz)
  call Sx_op(.false., .false., pertrho_pad, pertrhox)
  !Staggering on y if not twoD problem
  if(.not. twoD) then
  	allocate(pertrhoy(nz,nx,ny))
  	call Sy_op(.false., .false., pertrho_pad, pertrhoy)
  endif
  
  
  if (wantwavefield) write(0,*) "#Program will be writing wavefield as requested#"
  
  do is = 1,ns
  	old = 0.; cur = 0.
  	sec_source_left = 0.; sec_source_right = 0.
  	scat_sou = 0.; tmp_scat = 0.
  	source_wavefield = 0.
	write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	!Scaling the source
	scaled_wavelet = wavelet_fine
	call Sou_scale_acoustic_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_acoustic_op(scaled_wavelet,bulk_pad/rho_pad) !Adjoint = Forward (input-output same variable)
	call system_clock(count_rate=cr)
	call system_clock(count1)
	write(0,*)"Propagating source wavefield..."
	if (twoD) then
		call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
	else
		call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
	endif
	do itw = 1, nw-1 !Forward
		it = (itw-1)*desample
		!Propagate source field
		do it2 = 1, desample
			call Prop_acoustic_step_op(.false., pold, pcur, pnew)
			!Adding source (perturbed and source wavefield)
			if (twoD) then
				call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			else
				call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			endif	
			!write(0,*)"source it",it+it2
			!Apply adjoint(fine to coarse) of linear interpolation
			source_wavefield(:,:,:,itw) = source_wavefield(:,:,:,itw) + pcur*interp_filter(it2,1)
			source_wavefield(:,:,:,itw+1) = source_wavefield(:,:,:,itw+1) + pcur*interp_filter(it2,2)	
			!Swapping pointers
			pold => pcur 
			pcur => pnew
			pnew => pold
		enddo
		!saving source wavefield
		if (wantwavefield) then
			wc=srite("wavefields",source_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		
		if (mod(itw,(nw/10))==0) then
        	write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	
	enddo
		!Last step of the source wavefield just saving source wavefield
		source_wavefield(:,:,:,nw) = source_wavefield(:,:,:,nw) + pcur*interp_filter(desample+1,2)
		if (wantwavefield) then
			wc=srite("wavefields",source_wavefield(:,:,:,nw),nz*nx*ny*4)
		endif
		write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	write(0,'(a)',advance="yes")""
	write(0,*)"End Propagation source wavefield"
	
	write(0,*)"Propagating scattered wavefield..."
	old = 0.; cur = 0.
	poldpert => old
  	pcurpert => cur
  	pnewpert => old
  	!Creating secondary source (first step)
  	!Scattering caused by bulk modulus perturbation
  	call sec_deriv_field_op(source_wavefield, tmp_scat, 1)
  	scat_sou(:,:,:) = -tmp_scat(:,:,:)*pertk_pad(:,:,:)
  	!Scattering caused by density perturbation
  	!z-axis scattering 
  	call Dpz_op(.false.,.false.,source_wavefield(:,:,:,1),tmp_scat(:,:,:))
  	call Dmz_op(.false.,.true.,tmp_scat(:,:,:)*pertrhoz(:,:,:),scat_sou(:,:,:))
  	!x-axis scattering 
  	call Dpx_op(.false.,.false.,source_wavefield(:,:,:,1),tmp_scat(:,:,:))
  	call Dmx_op(.false.,.true.,tmp_scat(:,:,:)*pertrhox(:,:,:),scat_sou(:,:,:))
  	!y-axis scattering (3D ONLY) 
  	if(.not. twoD) then
  		call Dpy_op(.false.,.false.,source_wavefield(:,:,:,1),tmp_scat(:,:,:))
  		call Dmy_op(.false.,.true.,tmp_scat(:,:,:)*pertrhoy(:,:,:),scat_sou(:,:,:))
  	endif
	!Scale secondary source
	call weight_sec_op(.false., .false., scat_sou(:,:,:)*sec_scale(:,:,:), psec_source_left(:,:,:))

	
  	do itw = 1, nw-1 !Scattered wavefield
		!Propagate scattered wavefield
		it = (itw-1)*desample
		
		!Creating secondary source (right side)
  		!Scattering caused by bulk modulus perturbation
  		call sec_deriv_field_op(source_wavefield, tmp_scat, itw+1)
	  	scat_sou(:,:,:) = -tmp_scat(:,:,:)*pertk_pad(:,:,:)
  		!Scattering caused by density perturbation
  		!z-axis scattering 
  		call Dpz_op(.false.,.false.,source_wavefield(:,:,:,itw+1),tmp_scat(:,:,:))
  		call Dmz_op(.false.,.true.,tmp_scat(:,:,:)*pertrhoz(:,:,:),scat_sou(:,:,:))
  		!x-axis scattering 
  		call Dpx_op(.false.,.false.,source_wavefield(:,:,:,itw+1),tmp_scat(:,:,:))
  		call Dmx_op(.false.,.true.,tmp_scat(:,:,:)*pertrhox(:,:,:),scat_sou(:,:,:))
  		!y-axis scattering (3D ONLY) 
  		if(.not. twoD) then
  			call Dpy_op(.false.,.false.,source_wavefield(:,:,:,itw+1),tmp_scat(:,:,:))
  			call Dmy_op(.false.,.true.,tmp_scat(:,:,:)*pertrhoy(:,:,:),scat_sou(:,:,:))
  		endif
		!Scale secondary source
		call weight_sec_op(.false., .false., scat_sou(:,:,:)*sec_scale(:,:,:), psec_source_right(:,:,:))
		
		do it2 = 1, desample
			call Prop_acoustic_step_op(.false., poldpert, pcurpert, pnewpert)
			!Adding source (perturbed and source wavefield)
			pnewpert(:,:,:) = pnewpert(:,:,:) + psec_source_left(:,:,:)*interp_filter(it2,1) + psec_source_right(:,:,:)*interp_filter(it2,2)
			if (twoD) then
				!Taking perturbed wavefield
				do ig = 1, ng
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			else
				!Taking perturbed wavefield
				do ig = 1,ng
					call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
					call Add_source3D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			endif
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		!saving scattered wavefield
		if (wantwavefield) then
			wc=srite("wavefield",pcurpert,nz*nx*ny*4)
		endif
		
		!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		
		if (mod(itw,(nw/10))==0) then
        	write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
	
	enddo
		!saving scattered wavefield
		if (wantwavefield) then
			wc=srite("wavefield",pcurpert,nz*nx*ny*4)
		endif
		write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	write(0,'(a)',advance="yes")""
	call system_clock(count2)
	write(0,*) "Propagation time: ",(count2-count1)/cr
	write(0,'(A44,I5)') " Writing receiver wavefield for source n°: ", is
	!Resample receivers' wavefield
	if (dw .le. dt) then
		do ig = 1,ng
			wc=srite("out",recfield_fine(:,ig),nw*4)
		enddo
	else
		do ig = 1,ng
			call resample_op(.true.,.false.,recfield,recfield_fine(:,ig))
			wc=srite("out",recfield,nw*4)
		enddo
	endif
   	write(0,*) ""
	
	

  enddo
  call Prop_acoustic_step_close()
  call weight_sec_close()
  
  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",ng)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Receiver number")
  call to_history("n3",ns)
  call to_history("o3",1.)
  call to_history("d3",1.)
  call to_history("label3","Source number")
  
  
  
  if (wantwavefield) then
  
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",nt/desample+1,"wavefield")
  		call to_history("o3",ow,"wavefield")
 	 	call to_history("d3",dt*desample,"wavefield")
  		call to_history("label3","Time [s]","wavefield")
  		call to_history("n4",ns,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Source number","wavefield")  	

		call to_history("n1",nz,"wavefields")
  		call to_history("o1",oz,"wavefields")
  		call to_history("d1",dz,"wavefields")
  		call to_history("label1","z [m]","wavefields")
		call to_history("n2",nx,"wavefields")
	  	call to_history("o2",ox,"wavefields")
  		call to_history("d2",dx,"wavefields")
  		call to_history("label2","x [m]","wavefields")
		call to_history("n3",nt/desample+1,"wavefields")
  		call to_history("o3",ow,"wavefields")
 	 	call to_history("d3",dt*desample,"wavefields")
  		call to_history("label3","Time [s]","wavefields")
  		call to_history("n4",ns,"wavefields")
  		call to_history("o4",1.,"wavefields")
  		call to_history("d4",1.,"wavefields")
  		call to_history("label4","Source number","wavefields")  

  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
		call to_history("n4",nt/desample+1,"wavefield")
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dt*desample,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")

		call to_history("n1",nz,"wavefields")
  		call to_history("o1",oz,"wavefields")
  		call to_history("d1",dz,"wavefields")
  		call to_history("label1","z [m]","wavefields")
		call to_history("n2",nx,"wavefields")
	  	call to_history("o2",ox,"wavefields")
  		call to_history("d2",dx,"wavefields")
  		call to_history("label2","x [m]","wavefields")
		call to_history("n3",ny,"wavefields")
	  	call to_history("o3",oy,"wavefields")
	  	call to_history("d3",dy,"wavefields")
  		call to_history("label3","y [m]","wavefields")
		call to_history("n4",nt/desample+1,"wavefields")
  		call to_history("o4",ow,"wavefields")
 	 	call to_history("d4",dt*desample,"wavefields")
  		call to_history("label4","Time [s]","wavefields")
  		call to_history("n5",ns,"wavefields")
  		call to_history("o5",1.,"wavefields")
  		call to_history("d5",1.,"wavefields")
  		call to_history("label5","Source number","wavefields")
  	endif
  endif
  write(0,*) "End propagation"
  

  !Deallocating memory
  deallocate(tmp_scat,scat_sou)
  if(allocated(pertrhoy))deallocate(pertrhoy)
  deallocate(pertrhox,pertrhoz)
  deallocate(interp_filter,source_wavefield)
  deallocate(cur,old)
  deallocate(sec_source_right,sec_source_left)
  deallocate(sec_scale)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(pertrho_pad,pertk_pad,rho_pad,bulk_pad)
  deallocate(wavelet)
  call sep_close()
  call exit (0)

end program Bornfast_acoustic_fw