!Module to compute staggered partial derivative in staggered sense (Dp,Dm) (8th order approximation)
!Model dimension (z,x,y) Input: wavefield for Dp/ staggered wavefield for Dm
!Data dimension (z,x,y) Output: staggered wavefield for Dp/ wavefield for Dm
!D(z,x,y) = D8pt_(p,m)_(z,x,y)(Model(z,x,y))
!Model(z,x,y) = D8pt_(p,m)_(z,x,y)*(D(z,x,y))
module D8pt_mod

  implicit none
  !Module variables
  integer, private									:: nz, nx, ny
  real, private										:: dz, dx, dy
  logical, private									:: twoD
  real, private										:: C1z, C2z, C3z, C4z,C1x, C2x, C3x, C4x, C1y, C2y, C3y, C4y
  real, dimension(:,:,:),allocatable,private		:: tmp
  
  contains
  
  		subroutine D8pt_init(nz_in,nx_in,ny_in,dz_in,dx_in,dy_in,twoD_in)
			!Input variables	
			integer, intent(in)						:: nz_in, nx_in, ny_in
			real, intent(in)						:: dz_in, dx_in, dy_in
			logical, intent(in)						:: twoD_in
			!Internal variables
			integer 								:: buffer
			
			parameter(buffer=4)
			
			!Assigning module's variables
			nz = nz_in
			nx = nx_in
			ny = ny_in
			
			dz = dz_in
			dx = dx_in
			dy = dy_in
			
			twoD = twoD_in
			!Dp coefficients application --> C1(tmp(i+1)-tmp(i))+C2(tmp(i+2)-tmp(i-1))+C3(tmp(i+3)-tmp(i-2))......
			!Dm coefficients application --> C1(tmp(i)-tmp(i-1))+C2(tmp(i+1)-tmp(i-2))+C3(tmp(i+2)-tmp(i-3))......
			!Defining Dm coefficients
			C1z = 1.196289062541883/(dz)
			C2z = -0.079752604188901/(dz)
			C3z = 0.009570312506634/(dz)
			C4z = -6.975446437140719e-04/(dz)
		
			C1x = 1.196289062541883/(dx)
			C2x = -0.079752604188901/(dx)
			C3x = 0.009570312506634/(dx)
			C4x = -6.975446437140719e-04/(dx)

			C1y = 1.196289062541883/(dy)
			C2y = -0.079752604188901/(dy)
			C3y = 0.009570312506634/(dy)
			C4y = -6.975446437140719e-04/(dy)
		
			!Allocating memory for temporary array
			if (.not. allocated(tmp)) then
				if (twoD) then
					allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1))
				else
					allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1-buffer:ny+buffer))
				endif
			endif
			tmp = 0.
		
			return
		end subroutine D8pt_init
		
		
		!###############################################
		!Dpz partial derivative forward along z axis
		!###############################################
		subroutine Dpz_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Dpz_op(adj,add,mm,dd)
  		end subroutine Dpz_call
  		
  		
  		
  		subroutine Dpz_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
        		if(twoD) then
					!Applying (Dpz*)
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			mm(iz,ix,1) = mm(iz,ix,1) - (C1z*(tmp(iz,ix,1)-tmp(iz-1,ix,1))  &
										      		  + C2z*(tmp(iz+1,ix,1)-tmp(iz-2,ix,1)) &
										      		  + C3z*(tmp(iz+2,ix,1)-tmp(iz-3,ix,1)) &
										      		  + C4z*(tmp(iz+3,ix,1)-tmp(iz-4,ix,1)))
		           		enddo
		     		enddo
		        	!$OMP END PARALLEL DO
				
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       					    		do iz=1,nz
       					    			mm(iz,ix,iy) = mm(iz,ix,iy) - (C1z*(tmp(iz,ix,iy)-tmp(iz-1,ix,iy))  &
										      		 			  	+ C2z*(tmp(iz+1,ix,iy)-tmp(iz-2,ix,iy)) &
										      		  				+ C3z*(tmp(iz+2,ix,iy)-tmp(iz-3,ix,iy)) &
										      		  				+ C4z*(tmp(iz+3,ix,iy)-tmp(iz-4,ix,iy)))
       	     							
            						enddo
           					enddo
				 	enddo
        	 		!$OMP END PARALLEL DO
				
				endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				if(twoD) then
					!Applying Dzp
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			dd(iz,ix,1) = dd(iz,ix,1) + C1z*(tmp(iz+1,ix,1)-tmp(iz,ix,1))   &
										    		  + C2z*(tmp(iz+2,ix,1)-tmp(iz-1,ix,1)) &
										      		  + C3z*(tmp(iz+3,ix,1)-tmp(iz-2,ix,1)) &
										      		  + C4z*(tmp(iz+4,ix,1)-tmp(iz-3,ix,1)) 
		        	    enddo
		     	   enddo
		      	  !$OMP END PARALLEL DO
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       				    			do iz=1,nz
       	     							dd(iz,ix,iy) = dd(iz,ix,iy) + C1z*(tmp(iz+1,ix,iy)-tmp(iz,ix,iy))   &
										    		 			  	+ C2z*(tmp(iz+2,ix,iy)-tmp(iz-1,ix,iy)) &
										      		      		  	+ C3z*(tmp(iz+3,ix,iy)-tmp(iz-2,ix,iy)) &
										      		  			  	+ C4z*(tmp(iz+4,ix,iy)-tmp(iz-3,ix,iy)) 
            						enddo
           					enddo
			 		enddo
        	 		!$OMP END PARALLEL DO
				
				endif
		
			endif
   			
   			
   		end subroutine Dpz_op
   		
   		
   		!###############################################
		!Dpx partial derivative forward along x axis
		!###############################################
		subroutine Dpx_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Dpx_op(adj,add,mm,dd)
  		end subroutine Dpx_call
  		
  		
  		
  		subroutine Dpx_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
        		if(twoD) then
					!Applying (Dpx*)
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			mm(iz,ix,1) = mm(iz,ix,1)  -(C1x*(tmp(iz,ix,1)-tmp(iz,ix-1,1))   &
									     			   + C2x*(tmp(iz,ix+1,1)-tmp(iz,ix-2,1)) &
									 			       + C3x*(tmp(iz,ix+2,1)-tmp(iz,ix-3,1)) &
									 			       + C4x*(tmp(iz,ix+3,1)-tmp(iz,ix-4,1)))
		           		enddo
		     		enddo
		        	!$OMP END PARALLEL DO
				
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       					    		do iz=1,nz
       					    			mm(iz,ix,iy) = mm(iz,ix,iy) -(C1x*(tmp(iz,ix,iy)-tmp(iz,ix-1,iy))   &
									   	   					    	+ C2x*(tmp(iz,ix+1,iy)-tmp(iz,ix-2,iy)) &
									      	    					+ C3x*(tmp(iz,ix+2,iy)-tmp(iz,ix-3,iy)) &
									      	    					+ C4x*(tmp(iz,ix+3,iy)-tmp(iz,ix-4,iy))) 
            						enddo
           					enddo
				 	enddo
        	 		!$OMP END PARALLEL DO
				
				endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				if(twoD) then
					!Applying Dxp
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			dd(iz,ix,1) = dd(iz,ix,1) +	C1x*(tmp(iz,ix+1,1)-tmp(iz,ix,1))   &
									 		      	  + C2x*(tmp(iz,ix+2,1)-tmp(iz,ix-1,1)) &
									      		  	  + C3x*(tmp(iz,ix+3,1)-tmp(iz,ix-2,1)) &
									      		  	  + C4x*(tmp(iz,ix+4,1)-tmp(iz,ix-3,1))
		        	    enddo
		     	   enddo
		      	  !$OMP END PARALLEL DO
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       				    			do iz=1,nz
       	     							dd(iz,ix,iy) = 	dd(iz,ix,iy) + C1x*(tmp(iz,ix+1,iy)-tmp(iz,ix,iy))   &
											   						 + C2x*(tmp(iz,ix+2,iy)-tmp(iz,ix-1,iy)) &
																	 + C3x*(tmp(iz,ix+3,iy)-tmp(iz,ix-2,iy)) &
																     + C4x*(tmp(iz,ix+4,iy)-tmp(iz,ix-3,iy)) 
            						enddo
           					enddo
			 		enddo
        	 		!$OMP END PARALLEL DO
				
				endif
		
			endif
   			
   			
   		end subroutine Dpx_op
   		
   		!###############################################
		!Dpy partial derivative forward along y axis (3D ONLY)
		!###############################################
		subroutine Dpy_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Dpy_op(adj,add,mm,dd)
  		end subroutine Dpy_call
  		
  		
  		
  		subroutine Dpy_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       					    		do iz=1,nz
       					    			mm(iz,ix,iy) = mm(iz,ix,iy) -(C1y*(tmp(iz,ix,iy)-tmp(iz,ix,iy-1))   &
									 			  	    			+ C2y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy-2)) &
									   			   	    			+ C3y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-3)) &
									   			   	    			+ C4y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-4)))

            						enddo
           					enddo
				 	enddo
        	 		!$OMP END PARALLEL DO

			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       				    			do iz=1,nz
       	     							dd(iz,ix,iy) =  dd(iz,ix,iy) + C1y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy))   &
											   						 + C2y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-1)) &
																	 + C3y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-2)) &
																	 + C4y*(tmp(iz,ix,iy+4)-tmp(iz,ix,iy-3))
            						enddo
           					enddo
			 		enddo
        	 		!$OMP END PARALLEL DO
		
			endif
   			
   			
   		end subroutine Dpy_op
   		
   		
   		!###############################################
		!Dmz partial derivative backward along z axis
		!###############################################
		subroutine Dmz_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Dmz_op(adj,add,mm,dd)
  		end subroutine Dmz_call
  		
  		
  		
  		subroutine Dmz_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
        		if(twoD) then
					!Applying (Dmz*)
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			mm(iz,ix,1) = mm(iz,ix,1) -(C1z*(tmp(iz+1,ix,1)-tmp(iz,ix,1))   &
										      		  + C2z*(tmp(iz+2,ix,1)-tmp(iz-1,ix,1)) &
										      	  	  + C3z*(tmp(iz+3,ix,1)-tmp(iz-2,ix,1)) &
										      		  + C4z*(tmp(iz+4,ix,1)-tmp(iz-3,ix,1)))
		           		enddo
		     		enddo
		        	!$OMP END PARALLEL DO
				
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       					    		do iz=1,nz
       					    			mm(iz,ix,iy) =  mm(iz,ix,iy) -(C1z*(tmp(iz+1,ix,iy)-tmp(iz,ix,iy))   &
									   								 + C2z*(tmp(iz+2,ix,iy)-tmp(iz-1,ix,iy)) &
																	 + C3z*(tmp(iz+3,ix,iy)-tmp(iz-2,ix,iy)) &
																	 + C4z*(tmp(iz+4,ix,iy)-tmp(iz-3,ix,iy))) 
       	     							
            						enddo
           					enddo
				 	enddo
        	 		!$OMP END PARALLEL DO
				
				endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				if(twoD) then
					!Applying Dmz
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			dd(iz,ix,1) = dd(iz,ix,1) + C1z*(tmp(iz,ix,1)-tmp(iz-1,ix,1))   &
												      + C2z*(tmp(iz+1,ix,1)-tmp(iz-2,ix,1)) &
												      + C3z*(tmp(iz+2,ix,1)-tmp(iz-3,ix,1)) &
												      + C4z*(tmp(iz+3,ix,1)-tmp(iz-4,ix,1)) 
		        	    enddo
		     	   enddo
		      	  !$OMP END PARALLEL DO
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       				    			do iz=1,nz
       	     							dd(iz,ix,iy) = dd(iz,ix,iy) + C1z*(tmp(iz,ix,iy)-tmp(iz-1,ix,iy))   &
									   	   	    					+ C2z*(tmp(iz+1,ix,iy)-tmp(iz-2,ix,iy)) &
									    		  	    			+ C3z*(tmp(iz+2,ix,iy)-tmp(iz-3,ix,iy)) &
									      	   						+ C4z*(tmp(iz+3,ix,iy)-tmp(iz-4,ix,iy))
            						enddo
           					enddo
			 		enddo
        	 		!$OMP END PARALLEL DO
				
				endif
		
			endif
   			
   			
   		end subroutine Dmz_op
   		
   		
   		!###############################################
		!Dmx partial derivative backward along x axis
		!###############################################
		subroutine Dmx_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Dmx_op(adj,add,mm,dd)
  		end subroutine Dmx_call
  		
  		
  		
  		subroutine Dmx_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
        		if(twoD) then
					!Applying (Dmx*)
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			mm(iz,ix,1) = mm(iz,ix,1) -(C1x*(tmp(iz,ix+1,1)-tmp(iz,ix,1))   &
									 		      	  + C2x*(tmp(iz,ix+2,1)-tmp(iz,ix-1,1)) &
									      		  	  + C3x*(tmp(iz,ix+3,1)-tmp(iz,ix-2,1)) &
									      		  	  + C4x*(tmp(iz,ix+4,1)-tmp(iz,ix-3,1))) 
		           		enddo
		     		enddo
		        	!$OMP END PARALLEL DO
				
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       					    		do iz=1,nz
       	     							mm(iz,ix,iy) = mm(iz,ix,iy)	-(C1x*(tmp(iz,ix+1,iy)-tmp(iz,ix,iy))   &
											   						+ C2x*(tmp(iz,ix+2,iy)-tmp(iz,ix-1,iy)) &
																	+ C3x*(tmp(iz,ix+3,iy)-tmp(iz,ix-2,iy)) &
																	+ C4x*(tmp(iz,ix+4,iy)-tmp(iz,ix-3,iy))) 
            						enddo
           					enddo
				 	enddo
        	 		!$OMP END PARALLEL DO
				
				endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				if(twoD) then
					!Applying Dmx
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			dd(iz,ix,1) = dd(iz,ix,1) + C1x*(tmp(iz,ix,1)-tmp(iz,ix-1,1))   &
									     			  + C2x*(tmp(iz,ix+1,1)-tmp(iz,ix-2,1)) &
									 			      + C3x*(tmp(iz,ix+2,1)-tmp(iz,ix-3,1)) &
									 			      + C4x*(tmp(iz,ix+3,1)-tmp(iz,ix-4,1)) 
		        	    enddo
		     	   enddo
		      	  !$OMP END PARALLEL DO
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       				    			do iz=1,nz
       	     							dd(iz,ix,iy) = dd(iz,ix,iy) + C1x*(tmp(iz,ix,iy)-tmp(iz,ix-1,iy))   &
									   	   					    	+ C2x*(tmp(iz,ix+1,iy)-tmp(iz,ix-2,iy)) &
									      	    					+ C3x*(tmp(iz,ix+2,iy)-tmp(iz,ix-3,iy)) &
									      	    					+ C4x*(tmp(iz,ix+3,iy)-tmp(iz,ix-4,iy)) 
            						enddo
           					enddo
			 		enddo
        	 		!$OMP END PARALLEL DO
				
				endif
		
			endif
   			
   			
   		end subroutine Dmx_op
   		
   		!###############################################
		!Dmy partial derivative backward along y axis (3D ONLY)
		!###############################################
		subroutine Dmy_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Dmy_op(adj,add,mm,dd)
  		end subroutine Dmy_call
  		
  		
  		
  		subroutine Dmy_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       				    		do iz=1,nz
       	     						mm(iz,ix,iy) = mm(iz,ix,iy)	-(C1y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy))   &
											   					+ C2y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-1)) &
																+ C3y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-2)) &
																+ C4y*(tmp(iz,ix,iy+4)-tmp(iz,ix,iy-3)))
            					enddo
           				enddo
			 	enddo
         		!$OMP END PARALLEL DO

			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       			    			do iz=1,nz
       	     						dd(iz,ix,iy) = dd(iz,ix,iy) + C1y*(tmp(iz,ix,iy)-tmp(iz,ix,iy-1))   &
									 			  	    		+ C2y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy-2)) &
									   			   	    		+ C3y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-3)) &
									   			   	    		+ C4y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-4))
           						enddo
        				enddo
		 		enddo
        	 	!$OMP END PARALLEL DO
				
			endif

   			
   			
   		end subroutine Dmy_op
   		
   		
   		
   		
  		
  		
  		
		
		
		subroutine D8pt_close()
			if(allocated(tmp))deallocate(tmp)
			return
		end subroutine D8pt_close


end module D8pt_mod