!!$=head1 NAME
!!$
!!$Propagator Adjoint - Full acoustic propagator (i.e., varying density).
!!$
!!$=head1 SYNOPSIS
!!$
!!$Progator_acoustic_adj <bulk.H rho=rho.H data=data.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefield=wavefield.H >backdata.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation solver for full acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameter files created by Velpar.x and back propagates the data into
!!$the model and return the back propagated backdata.H
!!$NOTE: the data has to be sorted as common-source gather (same order as the input_recs.p and input_souces.p)
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<bulk.H> - Bulk modulus model in SEP3D format (theoretically [GPa]).
!!$
!!$=item from par
!!$
!!$B<filename> I<rho> - density model in SEP3D format (theoretically [Kg/m^3]).
!!$
!!$B<filename> I<data> - [data.H] Data to be back propagated.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional) NB!mod(dt_wavelet/dt_wavefield)=0
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, Pad_interp3D, Pad_interp3D, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Propagator_acoustic_adj
  
  !Used modules 
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use Add_source2D
  use Add_source3D
  use resample
  use fourier
  use Sou_scale_acoustic
  use Prop_acoustic_step


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: bulk, rho  !bulk modulus and density models 
  real, dimension(:), allocatable						:: wavelet !Source wavelet 
  integer												:: wantwavefield
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold
  real, dimension(:,:,:), pointer						:: bulk_pad, rho_pad !Padded models
  real													:: dt !Temporal sampling of the propagator
  real													:: maxvel, minvel !max and min velocity
  real, dimension(:), allocatable						:: wavelet_fine !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the velocity 
  integer												:: is, ig, it !Cycles variables
  integer												:: wc, srite, rc, sreed, sk, sseek
  integer												:: desample, icount
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr
  integer												:: count1, count2,ierr
  !Output variables 


  parameter( nv_pad=100) !Padding for the absording boundary

  call sep_init (SOURCE)

  write(0,*)""
  write(0,*)""
  write(0,*)"##############################################################################"
  write(0,*)"#################### Full Acoustic 2D/3D wave propagator #####################"
  write(0,*)"####################                Adjoint              #####################"  
  write(0,*)"##############################################################################"
  write(0,*)""
  write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("dt",dt,0.)
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz)
  call from_history("d1", dz)
  call from_history("n2", nx)
  call from_history("o2", ox)
  call from_history("d2", dx)
  call from_history("n3", ny)
  call from_history("o3", oy)
  call from_history("d3", dy)
  call from_aux ("data","n1", nw)
  call from_aux ("data","o1", ow)
  call from_aux ("data","d1", dw)

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Allocating the memory for back-propagated data
  allocate(wavelet(nw))
  wavelet = 0.

  !Reading the bulk modulus
  allocate(bulk(nz,nx,ny))
  bulk(:,:,:) = 0.
  rc=sreed("in" , bulk , nz*nx*ny*4)
  bulk(:,:,:) = bulk(:,:,:)*1e9 ! Conversion from GPa to Pa
  !Reading the density model
  allocate(rho(nz,nx,ny))
  rho(:,:,:) = 0.
  rc=sreed("rho" , rho , nz*nx*ny*4)
  !Computing max and min velocities
  minvel = minval(sqrt(bulk/rho))
  maxvel = maxval(sqrt(bulk/rho))

  !Adding absorbing boundaries to the bulk model
  call Pad_velocity(bulk,bulk_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  !Adding absorbing boundaries to the density model
  call Pad_perturbation(rho,rho_pad,nv_pad)
  
  write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  write(0,*) "	Number of samples for 50 Hz wave = ", (minvel/(50.*max(dx,dy,dz)))
  write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  write(0,*)"	Min vel [m/s]:",minvel
  write(0,*)"	Max vel [m/s]:",maxvel
  write(0,*) ""
  
  

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt))
  	 wavelet_fine = 0.
  else
 	nt = (nw-1)*desample+1
  	!Array for taking wavefield at the source position
	allocate(wavelet_fine(nt))
  	wavelet_fine = 0.
  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_acoustic_step_init(nz, nx, ny, nt, nv_pad, bulk_pad, rho_pad, dt, dz, dx, dy, twoD)
  call resample_init(nw, nt)
  

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  write(0,*)"Allocating memory for source and wavefield"
  !Allocating padded source and wavefield
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  old = 0.; cur = 0.
  pold => old
  pcur => cur
  pnew => old
  write(0,*) "Dimensions:"
  write(0,'(A7,I6)') " nz = ",nz
  write(0,'(A7,I6)') " nx = ",nx
  write(0,'(A7,I6)') " ny = ",ny
  write(0,'(A7,I6)') " nt = ",nt
  write(0,'(A7,I6)') " ns = ",ns
  write(0,'(A7,I6)') " ng = ",ng
  write(0,*) ""

  if (wantwavefield) then
	write(0,*) "#Program will be writing wavefield as requested#"
	do is = 1,ns
		do it = 1, nw
			wc=srite("wavefield",pcur,nz*nx*ny*4)
		enddo
	enddo
  endif
  

  do is = 1,ns
  	!moving the file writing position with sseek
  	if (wantwavefield) sk = sseek("wavefield",4*nz*nx*ny*nw*(is),0)
  	old = 0.; cur = 0.
  	write(0,'(A42,I5)') "Starting back propagation for source n°: ", is
	call system_clock(count_rate=cr) 
	call system_clock(count1)
  	!Creating adjoint source at the receivers
	if (dw .le. dt) then
		do ig = 1,ng
			rc=sreed("data",recfield,nw*4)
			recfield_fine(:,ig) = recfield 
		enddo
	else
		do ig = 1,ng
			rc=sreed("data",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_fine(:,ig))
		enddo
	endif
	write(0,*)"Propagating..."
	icount=1
	do it = nt, 1, -1 !Adjoint
		!Propagate
		call Prop_acoustic_step_op(.true., pold, pcur, pnew)
		!Adding adjoint source and Taking wavefield at source location
		if (twoD) then
			do ig = 1, ng
				call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
				call Add_source2D_op(.true.,.true.,pnew,recfield_fine(it,ig))
			enddo
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
			call Add_source2D_op(.false.,.false.,pcur,wavelet_fine(it))
		else
			do ig = 1, ng
				call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
				call Add_source3D_op(.true.,.true.,pnew,recfield_fine(it,ig))
			enddo
			call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
			call Add_source3D_op(.false.,.false.,pcur,wavelet_fine(it))
		endif
		if (wantwavefield) then
			if (mod(it,desample) .eq. 1) then
				sk = sseek("wavefield",-4*nz*nx*ny,1)
				wc=srite("wavefield",pcur,nz*nx*ny*4)
				sk = sseek("wavefield",-4*nz*nx*ny,1)
			endif
		endif
		if (mod(icount,(nt/10))==0) then
	        	write(0,'(I3,A4)',advance="no")nint((icount)*100./nt),"%..."
	        	close(0)
		endif		
		!Swapping pointers
		pold => pcur 
		pcur => pnew
		pnew => pold
		icount = icount + 1
	enddo
	write(0,'(a)',advance="yes")""
	call system_clock(count2)
	write(0,*) "Propagation time: ",(count2-count1)/cr
	!Scaling the backdata
	call Sou_scale_acoustic_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_acoustic_op(wavelet_fine,bulk_pad/rho_pad) !Adjoint = Forward (input-output same variable)
	write(0,'(A50,I5)') "  Writing back propagated data for source n°: ", is
	if (dw .le. dt) then
		wc=srite("out",real(wavelet_fine),nw*4)
	else
		call resample_op(.true.,.false.,wavelet,wavelet_fine)
		wc=srite("out",wavelet,nw*4) 
	endif
   	write(0,*) ""

  enddo

  !Closing used routines
  call Prop_acoustic_step_close()
  call Sou_scale_acoustic_close()

  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",ns)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Source number")
  if (wantwavefield) then
  
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",nt/desample+1,"wavefield")
  		call to_history("o3",ow,"wavefield")
 	 	call to_history("d3",dt*desample,"wavefield")
  		call to_history("label3","Time [s]","wavefield")
  		call to_history("n4",ns,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Source number","wavefield")  	
  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
		call to_history("n4",nt/desample+1,"wavefield")
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dt*desample,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")
  	endif
  endif
  write(0,*) "End propagation"


  !Deallocating memory
  deallocate(cur,old)
  deallocate(wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(rho_pad,bulk_pad)
  deallocate(wavelet)
  call sep_close()
  call exit (0)
end program Propagator_acoustic_adj
