!!$=head1 NAME
!!$
!!$Propagator Forward - Full acoustic propagator (i.e., varying density).
!!$
!!$=head1 SYNOPSIS
!!$
!!$Progator_acoustic_fw.x <bulk.H rho=rho.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefield=wavefield.H threads= >data.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation solver for full acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameter files created by Velpar.x and propagates the wavelet into
!!$the model and return the data.H
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<bulk.H> - Bulk modulus model in SEP3D format (theoretically [GPa]).
!!$
!!$=item from par
!!$
!!$B<filename> I<rho> - density model in SEP3D format (theoretically [Kg/m^3]).
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional) NB!mod(dt_wavelet/dt_wavefield)=0
!!$
!!$B<integer> I<threads> - [8] Number of threads to use during propagation.
!!$
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, Pad_interp3D, Pad_interp3D, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Propagator_acoustic_fw
  
  !Used modules 
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use Add_source2D
  use Add_source3D
  use resample 
  use fourier
  use Sou_scale_acoustic
  use Prop_acoustic_step


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: bulk, rho  !bulk modulus and density models 
  real, dimension(:), allocatable						:: wavelet, scaled_wavelet !Source wavelet 
  integer												:: wantwavefield, num_threads
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold
  real, dimension(:,:,:), pointer						:: bulk_pad, rho_pad !Padded models
  real													:: dt !Temporal sampling of the propagator
  real													:: maxvel, minvel !max and min velocity
  real, dimension(:), allocatable						:: wavelet_fine !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the models 
  integer												:: is, ig, it !Cycles variables
  integer												:: wc, srite, rc, sreed
  integer												:: desample
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr
  integer												:: count1, count2,ierr
  !Output variables 
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
  
		  

  write(0,*)""
  write(0,*)""
  write(0,*)"##############################################################################"
  write(0,*)"#################### Full Acoustic 2D/3D wave propagator #####################"
  write(0,*)"####################                Forward              #####################"  
  write(0,*)"##############################################################################"
  write(0,*)""
  write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("dt",dt,0.)
  
  !Setting number of threads
  call from_par("threads",num_threads,8)
  call omp_set_num_threads(num_threads)
  
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy);
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the bulk modulus
  allocate(bulk(nz,nx,ny))
  bulk(:,:,:) = 0.
  rc=sreed("in" , bulk , nz*nx*ny*4)
  bulk(:,:,:) = bulk(:,:,:)*1e9 ! Conversion from GPa to Pa
  !Reading the density model
  allocate(rho(nz,nx,ny))
  rho(:,:,:) = 0.
  rc=sreed("rho" , rho , nz*nx*ny*4)
  !Computing max and min velocities
  minvel = minval(sqrt(bulk/rho))
  maxvel = maxval(sqrt(bulk/rho))

  !Adding absorbing boundaries to the bulk model
  call Pad_velocity(bulk,bulk_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  !Adding absorbing boundaries to the density model
  call Pad_perturbation(rho,rho_pad,nv_pad)
  
  write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  write(0,*) "	Number of samples for 50 Hz wave = ", (minvel/(50.*max(dx,dy,dz)))
  write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  write(0,*)"	Min vel [m/s]:",minvel
  write(0,*)"	Max vel [m/s]:",maxvel
  write(0,*) ""
  
  

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	call resample_init(nw, nt)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)

  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_acoustic_step_init(nz, nx, ny, nt, nv_pad, bulk_pad, rho_pad, dt, dz, dx, dy, twoD)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  pold => old
  pcur => cur
  pnew => old
  write(0,*) "Dimensions:"
  write(0,'(A7,I6)') " nz = ",nz
  write(0,'(A7,I6)') " nx = ",nx
  write(0,'(A7,I6)') " ny = ",ny
  write(0,'(A7,I6)') " nt = ",nt
  write(0,'(A7,I6)') " ns = ",ns
  write(0,'(A7,I6)') " ng = ",ng
  write(0,*) ""
  
  if (wantwavefield) write(0,*) "#Program will be writing wavefield as requested#"

  do is = 1,ns
  	old = 0.; cur = 0.
	write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	!Scaling the source
	scaled_wavelet = wavelet_fine
	call Sou_scale_acoustic_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_acoustic_op(scaled_wavelet,bulk_pad/rho_pad) !Adjoint = Forward (input-output same variable)
	call system_clock(count_rate=cr)
	call system_clock(count1)
	write(0,*)"Propagating..."
	do it = 1, nt !Forward
		!Propagate
		call Prop_acoustic_step_op(.false., pold, pcur, pnew)
		!Adding source and Taking wavefield
		if (twoD) then
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
			call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it))
			do ig = 1, ng
				call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
				call Add_source2D_op(.false.,.false.,pcur,recfield_fine(it,ig))
			enddo
		else
			call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
			call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it))
			do ig = 1,ng
				call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
				call Add_source3D_op(.false.,.false.,pcur,recfield_fine(it,ig))
			enddo
		endif
		if (wantwavefield) then
			if (mod(it,desample) .eq. 1) then
				wc=srite("wavefield",pcur,nz*nx*ny*4)
			endif
		endif
		if (mod(it,(nt/10))==0) then
        	write(0,'(I3,A4)',advance="no")nint(it*100./nt),"%..."
        	close(0)
      	endif
		!Swapping pointers
		pold => pcur 
		pcur => pnew
		pnew => pold
	enddo
	write(0,'(a)',advance="yes")""
	call system_clock(count2)
	write(0,*) "Propagation time: ",(count2-count1)/cr
	write(0,'(A44,I5)') " Writing receiver wavefield for source n°: ", is
	!Resample receivers' wavefield
	if (dw .le. dt) then
		do ig = 1,ng
			wc=srite("out",recfield_fine(:,ig),nw*4)
		enddo
	else
		do ig = 1,ng
			call resample_op(.true.,.false.,recfield,recfield_fine(:,ig))
			wc=srite("out",recfield,nw*4)
		enddo
	endif
   	write(0,*) ""

  enddo

  !Closing used routines
  call Prop_acoustic_step_close()
  call Sou_scale_acoustic_close()

  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",ng)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Receiver number")
  call to_history("n3",ns)
  call to_history("o3",1.)
  call to_history("d3",1.)
  call to_history("label3","Source number")
  if (wantwavefield) then
  
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",nt/desample+1,"wavefield")
  		call to_history("o3",ow,"wavefield")
 	 	call to_history("d3",dt*desample,"wavefield")
  		call to_history("label3","Time [s]","wavefield")
  		call to_history("n4",ns,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Source number","wavefield")  	
  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
		call to_history("n4",nt/desample+1,"wavefield")
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dt*desample,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")
  	endif
  endif
  write(0,*) "End propagation"

  !Deallocating memory
  deallocate(cur,old)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(rho_pad,bulk_pad)
  deallocate(wavelet)
  call sep_close()
  call exit (0)
end program Propagator_acoustic_fw


