!Module to compute staggered m2(1/d_density) perturbation for Born/RTM in non-constant density approximation
!Model dimension (z,x,y) Input regular grid density
!Data dimension (z,x,y) Output staggered 1/density perturbation
!D(z,x,y) = S_(z,x,y)(Model(z,x,y))
!Model(z,x,y) = S_(z,x,y)*(D(z,x,y))
module S_mod

  implicit none
  !Module variables
  integer, private					:: nz, nx, ny
  logical, private					:: twoD

  contains
		
		subroutine S_init(nz_in,nx_in,ny_in,twoD_in)
			!Input variables	
			integer, intent(in)						:: nz_in, nx_in, ny_in
			logical, intent(in)						:: twoD_in
			!Internal variables
			
			
			nz = nz_in
			nx = nx_in
			ny = ny_in
			twoD = twoD_in
		
		
		end subroutine S_init
		
		!###############################################
		!Staggering along z axes
		!###############################################
		subroutine Sz_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Sz_op(adj,add,mm,dd)
  		end subroutine Sz_call
  		
  		subroutine Sz_op(adj, add, mm, dd)
	
			!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		
        		if(twoD) then
        			!Boundary condition of adjoint (beginning points)
        			mm(1,1:nx,1) = mm(1,1:nx,1) + 0.5*dd(1,1:nx,1)
        			!Bz m2 on (i,j,k)
        			!OMP PARALLEL DO private(ix,iz)
					do ix = 1, nx
						do iz = 2, nz-1
							mm(iz,ix,1) = mm(iz,ix,1) + 0.5*(dd(iz-1,ix,1) + dd(iz,ix,1))
						enddo
					enddo	
					!OMP END PARALLEL DO 
					!Boundary grid points
					mm(nz,1:nx,1) = mm(nz,1:nx,1) + dd(nz,1:nx,1) + 0.5*dd(nz-1,1:nx,1) !Last depth elements
        		
        		else
        			!Boundary condition of adjoint (beginning points)
        			mm(1,1:nx,1:ny) = mm(1,1:nx,1:ny) + 0.5*dd(1,1:nx,1:ny)
					!$OMP PARALLEL DO private(iy,ix,iz)
					do iy = 1, ny
						do ix = 1, nx
							do iz = 2, nz-1
								mm(iz,ix,iy) = mm(iz,ix,iy) + 0.5*(dd(iz-1,ix,iy) + dd(iz,ix,iy))
							enddo
						enddo
					enddo	
					!$OMP END PARALLEL DO
					!Boundary grid points
					mm(nz,1:nx,1:ny) = mm(nz,1:nx,1:ny) + dd(nz,1:nx,1:ny) + 0.5*dd(nz-1,1:nx,1:ny) !Last depth elements
        		
        		endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				
				if(twoD) then
        			!m2 on (i+1/2,j,1)
        			!$OMP PARALLEL DO private(ix,iz)
					do ix = 1, nx
						do iz = 1, nz-1
							dd(iz,ix,1) = dd(iz,ix,1) + 0.5*(mm(iz,ix,1) + mm(iz+1,ix,1))
						enddo
					enddo	
					!$OMP END PARALLEL DO 
					!Boundary grid points
					dd(nz,1:nx,1) = dd(nz,1:nx,1) + mm(nz,1:nx,1) !Last depth elements
        		else
        			!m2 on (i+1/2,j,k)
					!$OMP PARALLEL DO private(iy,ix,iz)
					do iy = 1, ny
						do ix = 1, nx
							do iz = 1, nz-1
								dd(iz,ix,iy) = dd(iz,ix,iy) + 0.5*(mm(iz,ix,iy) + mm(iz+1,ix,iy))
							enddo
						enddo
					enddo	
					!$OMP END PARALLEL DO
					!Boundary grid points
					dd(nz,1:nx,1:ny) = dd(nz,1:nx,1:ny) + mm(nz,1:nx,1:ny) !Last depth elements
        		endif

		
			endif
			
		end subroutine Sz_op
		
		!###############################################
		!Staggering along x axes
		!###############################################
		subroutine Sx_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Sx_op(adj,add,mm,dd)
  		end subroutine Sx_call
  		
  		subroutine Sx_op(adj, add, mm, dd)
	
			!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		
        		if(twoD) then
        			!Boundary condition of adjoint (beginning points)
        			mm(1:nz,1,1) = mm(1:nz,1,1) + 0.5*dd(1:nz,1,1)
        			!Bz m2 on (i,j,k)
        			!OMP PARALLEL DO private(ix,iz)
					do ix = 2, nx-1
						do iz = 1, nz
							mm(iz,ix,1) = mm(iz,ix,1) + 0.5*(dd(iz,ix-1,1) + dd(iz,ix,1))
						enddo
					enddo	
					!OMP END PARALLEL DO 
					!Boundary grid points
					mm(1:nz,nx,1) = mm(1:nz,nx,1) + dd(1:nz,nx,1) + 0.5*dd(1:nz,nx-1,1) !Last horizontal-x elements

        		else
        			!Boundary condition of adjoint (beginning points)
        			mm(1:nz,1,1:ny) = mm(1:nz,1,1:ny) + 0.5*dd(1:nz,1,1:ny)
					!$OMP PARALLEL DO private(iy,ix,iz)
					do iy = 1, ny
						do ix = 2, nx-1
							do iz = 1, nz
								mm(iz,ix,iy) = mm(iz,ix,iy) + 0.5*(dd(iz,ix-1,iy) + dd(iz,ix,iy))
							enddo
						enddo
					enddo	
					!$OMP END PARALLEL DO
					!Boundary grid points
					mm(1:nz,nx,1:ny) = mm(1:nz,nx,1:ny) + dd(1:nz,nx,1:ny) + 0.5*dd(1:nz,nx-1,1:ny) !Last horizontal-x elements
        		
        		endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				
				if(twoD) then
        			!m2 on (i,j+1/2,k)
					!$OMP PARALLEL DO private(ix,iz)
					do ix = 1, nx-1
						do iz = 1, nz
							dd(iz,ix,1) = dd(iz,ix,1) + 0.5*(mm(iz,ix,1) + mm(iz,ix+1,1))
						enddo
					enddo	
					!$OMP END PARALLEL DO
					!Boundary grid points
					dd(1:nz,nx,1) = dd(1:nz,nx,1) + mm(1:nz,nx,1) !Last horizontal-x elements
        		else
        			!m2 on (i,j+1/2,k)
					!$OMP PARALLEL DO private(iy,ix,iz)
					do iy = 1, ny
						do ix = 1, nx-1
							do iz = 1, nz
								dd(iz,ix,iy) = dd(iz,ix,iy) + 0.5*(mm(iz,ix,iy) + mm(iz,ix+1,iy))
							enddo
						enddo
					enddo	
					!$OMP END PARALLEL DO
					!Boundary grid points
					dd(1:nz,nx,1:ny) = dd(1:nz,nx,1:ny) + mm(1:nz,nx,1:ny) !Last horizontal-x elements
        		
        		endif

		
			endif
			
		end subroutine Sx_op
		
		
		!###############################################
		!Staggering along y axes (FOR 3D ONLY)
		!###############################################
		subroutine Sy_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call Sy_op(adj,add,mm,dd)
  		end subroutine Sy_call
  		
  		subroutine Sy_op(adj, add, mm, dd)
	
			!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		
        		
        		!Boundary condition of adjoint (beginning points)
        		mm(1:nz,1:nx,1) = mm(1:nz,1:nx,1) + 0.5*dd(1:nz,1:nx,1)
				!$OMP PARALLEL DO private(iy,ix,iz)
				do iy = 2, ny-1
					do ix = 1, nx
						do iz = 1, nz
							mm(iz,ix,iy) = mm(iz,ix,iy) + 0.5*(dd(iz,ix,iy-1) + dd(iz,ix,iy))
						enddo
					enddo
				enddo	
				!$OMP END PARALLEL DO
				!Boundary grid points
				mm(1:nz,1:nx,ny) = mm(1:nz,1:nx,ny) + dd(1:nz,1:nx,ny) + 0.5*dd(1:nz,1:nx,ny-1) !Last horizontal-y elements


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				
				!m2 on (i,j,k+1/2)
				!$OMP PARALLEL DO private(iy,ix,iz)
				do iy = 1, ny-1
					do ix = 1, nx
						do iz = 1, nz
							dd(iz,ix,iy) = dd(iz,ix,iy) + 0.5*(mm(iz,ix,iy) + mm(iz,ix,iy+1))
						enddo
					enddo
				enddo	
				!$OMP END PARALLEL DO
				!Boundary grid points
				dd(1:nz,1:nx,ny) = dd(1:nz,1:nx,ny) + mm(1:nz,1:nx,ny) !Last horizontal-y elements
		
			endif
			
		end subroutine Sy_op



end module S_mod