!!$=head1 NAME
!!$
!!$Velpar - Return the input parameters for Vel given the acquisition geometry taken from the headers with Geometry.sh
!!$
!!$=head1 SYNOPSIS
!!$
!!$Velpar source= recs= fmax=50. vmin=1500. s=4 maxdepth= o3=0. convsx=1 convsx=1 convsy=1 convsz=1 convgx=1 convgy=1 convgz=1 velfile=velpar.p origin=origin.p file_sources=input_sources.p file_recs=input_recs.p >/dev/null
!!$
!!$=head1 DESCRIPTION
!!$
!!$Return the input parameters for Vel (n1, n2, n3, o1, o2, o3, d1, d2, d3) given acquisition geometry, maximum frequency, minimum velocity,
!!$S(minimum number of grid points per wavelength).
!!$It also returns a source position file (input_sources.p) and a receiver position file (input_recs.p) with coordinates converted in meters.
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from par
!!$
!!$B<filename> I<source> - File containing x,y,z of the sources
!!$
!!$B<filename> I<recs> - File containing x,y,z of the receivers
!!$
!!$B<real> I<fmax> - [50.] Maximum frequency to be propagated [Hz]
!!$
!!$B<real> I<vmin> - [1500.] Minumim velocity present in the velocity model [m/s]
!!$
!!$B<real> I<maxdepth> - Maximum depth of the velocity model [m]
!!$
!!$B<real> I<o3> - [0.] Origin point of the z-axis [m]
!!$
!!$B<int> I<s> - [4] Minimum number of grid points per wavelength
!!$
!!$B<real> I<convsx,convsy,convsz,convsz,convgx,convgy,convgz> - [1.] Conversion factors to have coordinates in [m]eters (e.g. dm -> m = 0.1)
!!$
!!$B<filename> I<velfile> - [velpar.p] Output file with Vel parameters
!!$
!!$B<filename> I<origin> - [origin.p] Output file with model o1, o2, o3 (problem with Vel if o1,o2,o3 are big numbers)
!!$
!!$B<filename> I<file_sources> - [input_sources.p] Output file with the converted source positions
!!$
!!$B<filename> I<file_recs> - [input_recs.p] Output file with the converted receiver positions
!!$
!!$
!!$=back
!!$
!!$=head1 SEE ALSO
!!$
!!$Vel, Geometry.sh
!!$
!!$=cut
!%
program Velpar
  
  !Used modules 
  use sep

  !Variable declaration
  implicit none
  !Input variables
  character*256  		    			:: source, recs, velfile, path, sources_file, recs_file, origin
  real							:: fmax, vmin, convsx, convsy, convsz, convgx, convgy, convgz
  real							:: o3, maxdepth
  integer						:: s
  !Internal variables
  integer						:: open_stat,read_stat
  integer						:: n_source, n_rec, i
  character*256  		    			:: line_buffer
  character*15, dimension(7)				:: num_buffer
  real, dimension(:), allocatable			:: sx, sy, sz, gx, gy, gz
  real							:: max_x, max_y, max_x_rec, max_y_rec, max_x_sou, max_y_sou
  real							:: min_x_rec, min_y_rec, min_x_sou, min_y_sou
  real							:: ds !Model spatial sampling
  !Output variables
  integer						:: n1, n2, n3
  real							:: o1, o2

  call sep_init (SOURCE)

  !Reading parameters
  call from_par ("source",source)
  call from_par ("recs",recs)
  call from_par ("velfile",velfile,"velpar.p")
  call from_par ("origin",origin,"origin.p")
  call from_par ("file_sources",sources_file,"input_sources.p")
  call from_par ("file_recs",recs_file,"input_recs.p")
  call from_par ("fmax",fmax,50.)
  call from_par ("vmin",vmin,1500.)
  call from_par ("convsx",convsx,1.)
  call from_par ("convsy",convsy,1.)
  call from_par ("convsz",convsz,1.)
  call from_par ("convgx",convgx,1.)
  call from_par ("convgy",convgy,1.)
  call from_par ("convgz",convgz,1.)
  call from_par ("s",s,4)
  call from_par ("maxdepth",maxdepth)
  call from_par ("o3",o3,0.)

  !Reading and writing source positions
  open(file=source,unit=22,form='formatted', status='old', iostat=open_stat)
  open(file=sources_file,unit=24, status='unknown')
  if (open_stat .ne. 0 ) then
  	write(0,*) "Cannot read the input file for sources"
  	stop
  endif
  read_stat=0
  n_source=0
  do while (read_stat == 0)
  	read(unit=22,fmt='(A)',iostat=read_stat) line_buffer
  	if (read_stat == 0) then
  		if (line_buffer(1:1)=='#') cycle
	        n_source = n_source + 1
        	read(line_buffer,*) 
  	endif
  enddo
  write(24,*) "# Number of sources"
  write(24,*) n_source
  write(24,*) "# Source positions (x,y,z)[m]"
  rewind(unit=22)
  allocate(sx(n_source),sy(n_source),sz(n_source))
  read_stat=0
  i=0
  do while (read_stat == 0)
  	read(unit=22,fmt='(A)',iostat=read_stat) line_buffer
  	if (read_stat == 0) then
  		if (line_buffer(1:1)=='#') cycle
       		i = i + 1
        	read(line_buffer,*) sx(i), sy(i), sz(i)
		sx(i) = sx(i)*convsx
		sy(i) = sy(i)*convsy
		sz(i) = sz(i)*convsz
		write(24,*) sx(i),sy(i),sz(i)
  	endif
  enddo

  !Reading and writing receiver positions
  open(file=recs,unit=23,form='formatted', status='old', iostat=open_stat)
  open(file=recs_file,unit=25, status='unknown')
  if (open_stat .ne. 0 ) then
  	write(0,*) "Cannot read the input file for receivers"
  	stop
  endif
  read_stat=0
  n_rec=0
  do while (read_stat == 0)
  	read(unit=23,fmt='(A)',iostat=read_stat) line_buffer
  	if (read_stat == 0) then
  		if (line_buffer(1:1)=='#') cycle
	        n_rec = n_rec + 1
        	read(line_buffer,*) 
  	endif
  enddo
  write(25,*) "# Number of receivers"
  write(25,*) n_rec
  write(25,*) "# Receivers positions (x,y,z)[m]"
  rewind(unit=23)
  allocate(gx(n_rec),gy(n_rec),gz(n_rec))
  read_stat=0
  i=0
  do while (read_stat == 0)
  	read(unit=23,fmt='(A)',iostat=read_stat) line_buffer
  	if (read_stat == 0) then
  		if (line_buffer(1:1)=='#') cycle
       		i = i + 1
        	read(line_buffer,*) gx(i), gy(i), gz(i)
		gx(i) = gx(i)*convgx
		gy(i) = gy(i)*convgy
		gz(i) = gz(i)*convgz
		write(25,*) gx(i),gy(i),gz(i)
  	endif
  enddo
  
  !Find the limits of the Velocity model (x,y)
  max_x_rec = maxval(gx(:))
  max_x_sou = maxval(sx(:))
  !max x-limit
  max_x = max(max_x_rec,max_x_sou)
  max_y_rec = maxval(gy(:))
  max_y_sou = maxval(sy(:))
  !max y-limit
  max_y = max(max_y_rec,max_y_sou)
  min_x_rec = minval(gx(:))
  min_x_sou = minval(sx(:))
  !x-origin
  o1 = min(min_x_rec,min_x_sou)
  min_y_rec = minval(gy(:))
  min_y_sou = minval(sy(:))
  !y-origin
  o2 = min(min_y_rec,min_y_sou)
  !Computing the maximum spatial sampling (for avoiding numerical dispersion)
  ds = (vmin/(fmax*s))
  !Computing number of samples for each dimension
  n1 = 1.5 + (max_x - o1)/ds
  n2 = 1.5 + (max_y - o2)/ds
  maxdepth = max(maxdepth,maxval(gz(:)))
  n3 = 1.5 + (maxdepth - o3)/ds


!  Writing Vel parameter output files
  open(file=velfile,unit=26, status='unknown', iostat=open_stat)
  if (open_stat .ne. 0 ) then
  	write(0,*) "Cannot open the output file for Vel"
  	stop
  endif
  open(file=origin,unit=27, status='unknown', iostat=open_stat)
  if (open_stat .ne. 0 ) then
  	write(0,*) "Cannot open the output file for Vel"
  	stop
  endif

  write(num_buffer(1),'(I6)') n3 !n1 = depth
  write(num_buffer(2),'(I6)') n1 !n2 = x-axis
  write(num_buffer(3),'(I6)') n2 !n3 = y-axis
  write(num_buffer(4),'(F15.3)') o3 !o1 = depth
  write(num_buffer(5),'(F15.3)') o1 !o2 = x-axis
  write(num_buffer(6),'(F15.3)') o2 !o3 = y-axis
  write(num_buffer(7),'(F6.3)') ds
  write(26,*)"n1=",trim(adjustl(num_buffer(1)))," n2=",trim(adjustl(num_buffer(2)))," n3=",trim(adjustl(num_buffer(3))), &
	     " d1=",trim(adjustl(num_buffer(7)))," d2=",trim(adjustl(num_buffer(7)))," d3=",trim(adjustl(num_buffer(7))), &
	     " o1=",trim(adjustl(num_buffer(4)))," o2=",trim(adjustl(num_buffer(5)))," o3=",trim(adjustl(num_buffer(6)))

  write(27,*)" o1=",trim(adjustl(num_buffer(4)))," o2=",trim(adjustl(num_buffer(5)))," o3=",trim(adjustl(num_buffer(6)))

  deallocate(gz,gy,gx,sz,sy,sx)
  close(22);close(23);close(24);close(25);close(26);close(27)
  call sep_close()
  call exit (0)
end program Velpar
